//  MidiInterface
//  Adapted from Simple mmj demo (Nils Peters 1/13/07).
//  Copyright (c) 2007 __humatic__. All rights reserved.
//  Modified by Matthieu Quignard on 11/15/10

package nova;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import java.util.Vector;
import java.util.concurrent.*;

/* Generic class doing the interface between the editor (NovaManager) and the Nova System through Midi
   -- receive data, do some checking and trigger relevant method from NovaManager
   -- send data (called from NovaManager)
   -- user can choose midi ports and check whether a Nova is found on the chain 
      (respond to Universal Device Inquiry messages)

   Known subclasses
   -- MidiDefaultInterface : use javax.sound.midi
   -- MidiMacInterface : use javax.sound.midi or de.humatic.mmj package to work around bugs from Apple Midi implementation.
 */
public abstract class MidiInterface extends JFrame implements ItemListener, Runnable {

    Integer id = null; // SysExID of the Current NovaSystem. Null if standalone
    JComboBox ids;    // NovaList

    boolean enableMMJ = false;
    boolean useMMJ = false;

    JCheckBox cbMMJ; // CheckBox for choosing Midi API
    JComboBox ib; // List of Inputs
    JComboBox ob; // List of Outputs
    JLabel resultLabel; //displays whether Nova System have been found
    
    NovaManager parent;

    boolean inquire = false;

    Vector<Integer> novas = new Vector<Integer>();

    ConcurrentLinkedQueue<MMidiMessage> queue;

    static byte[] inqMessage = new byte[] { (byte) 0xF0, // Start SysEx
					    (byte) 0x7E, // Non Real Time
					    (byte) 0x7F, // Device Id [2] ALL
					    (byte) 0x06, 
					    (byte) 0x01, 
					    (byte) 0xF7 // End SysEx
    };

    // F0 7E 01 06 02 00 20 1F 63 00 00 00 00 00 01 0D F7
    static byte[] meNova = new byte[] { (byte) 0xF0, // Start SysEx
					(byte) 0x7E, // Non Real Time
					(byte) 0x00,// Device ID [2]
					(byte) 0x06, 
					(byte) 0x02, 
					(byte) 0x00, (byte) 0x20, (byte) 0x1F, // TC Electronic
					(byte) 0x63, //                           Nova System
					(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00, 
					(byte) 0x01, (byte) 0x0D, //              Nova System Software Version
					(byte) 0xF7
    };

    static byte[] reqPreset = new byte[] { (byte) 0xF0,
					   (byte) 0x00, (byte) 0x20, (byte) 0x1F, // TC Electronic
					   (byte) 0x00, // Device Id [4]
					   (byte) 0x63, // Nova System
					   (byte) 0x45, // Request
					   (byte) 0x01, // Data Requested is Preset
					   (byte) 0x00, // Preset number [8]
					   (byte) 0x00, // Void
					   (byte) 0xF7
    };

    static byte[] aPreset = new byte[] { (byte) 0xF0,
					 (byte) 0x00, (byte) 0x20, (byte) 0x1F, // TC Electronic
					 (byte) 0x00, // Device Id [4]
					 (byte) 0x63, // Nova System
					 (byte) 0x20, // It's a dump
					 (byte) 0x01  // It's a preset
    };

    // Preset Number Out of Bounds or preset Empty : F0 7E (ID) 7E 00 F7 
    static byte[] oob = new byte[] { (byte) 0xF0, // Start SysEx
				     (byte) 0x7E, // Non Real Time
				     (byte) 0x00, // Device Id [4]
				     (byte) 0x7E, // NAK
				     (byte) 0x00, // void
				     (byte) 0xF7  // End SysEx
    };

    static byte[] reqSystem = new byte[] { (byte) 0xF0,
					   (byte) 0x00, (byte) 0x20, (byte) 0x1F, // TC Electronic
					   (byte) 0x00, // Device Id [4]
					   (byte) 0x63, // Nova System
					   (byte) 0x45, // Request
					   (byte) 0x02, // Data Requested is System
					   (byte) 0x00, // Void
					   (byte) 0x00, // Void
					   (byte) 0xF7
    };

    static byte[] aSystem = new byte[] { (byte) 0xF0,
					 (byte) 0x00, (byte) 0x20, (byte) 0x1F, // TC Electronic
					 (byte) 0x00, // Device Id [4]
					 (byte) 0x63, // Nova System
					 (byte) 0x20, // It's a dump
					 (byte) 0x02  // This is system data
    };

    public MidiInterface() {
	super("Midi Connections");
	this.setDefaultCloseOperation(javax.swing.WindowConstants.HIDE_ON_CLOSE);
	parent = null;

	queue = new ConcurrentLinkedQueue<MMidiMessage>();

	ib = new JComboBox();
	ob = new JComboBox();
	resultLabel = new JLabel("Can't find any Nova System on this chain.", JLabel.CENTER);

	init();
	buildInterface();
	pack();
	ib.setSelectedIndex(ib.getItemCount()-1);
	ob.setSelectedIndex(ib.getItemCount()-1);
    }

    public MidiInterface(NovaManager np) {
	super("Midi Connections");
	this.setDefaultCloseOperation(javax.swing.WindowConstants.HIDE_ON_CLOSE);
	this.parent = np;

	queue = new ConcurrentLinkedQueue<MMidiMessage>();

	ib = new JComboBox();
	ob = new JComboBox();
	resultLabel = new JLabel("Can't find any Nova System on this chain.", JLabel.CENTER);
	ids = new JComboBox();
	
	init();//abstract : preprocessing
	buildInterface(); 
	setComboListeners(); //abstract: post-processing.
	pack();
	
	ib.setSelectedIndex(ib.getItemCount()-1);
	ob.setSelectedIndex(ib.getItemCount()-1);
    }
    
    // ABSTRACT METHODS

    abstract void init();
    abstract void rebuildCombos();
    abstract void setComboListeners();
    abstract void send(byte[] data);

    // END OF ABSTRACT METHODS


    protected void buildInterface() {
	boolean sendActiveSensing = false;

	cbMMJ = new JCheckBox( "Use MMJ instead of Javax.sound.midi API", useMMJ);
	cbMMJ.addItemListener(this);
	cbMMJ.setEnabled(enableMMJ);
			   
	resultLabel.setMinimumSize( new Dimension(100,100) );
	resultLabel.setText("Can't find any Nova System on this chain.");
	resultLabel.setFont(new Font("SansSerif", Font.ITALIC, 14));
	
	rebuildCombos();// abstract
	
	JButton ok = new JButton("Close");
	ok.addActionListener(new ActionListener(){
		public void actionPerformed(ActionEvent e) {
		    // refresh NovaManager Menus
		    setVisible( false );
		}
	    });

	this.getContentPane().setLayout(new BorderLayout());
	JPanel p = new JPanel();
	p.setLayout(new BorderLayout());

	JPanel l1 = new JPanel();
	if (enableMMJ) l1.add(cbMMJ);

	JPanel l2 = new JPanel(new GridLayout(2,2));
	l2.add(new JLabel("Select Midi input port", JLabel.CENTER));
	l2.add(new JLabel("Select Midi output port", JLabel.CENTER));
	l2.add(ib);
	l2.add(ob);   
	p.add("North", l1);
	p.add("Center", l2);

	JPanel l3 = new JPanel();
	l3.add(ok);

	JPanel lcentre = new JPanel();
	lcentre.add(resultLabel);
	ids.setVisible(false);
	lcentre.add(ids);

	this.getContentPane().add("North", p);
	this.getContentPane().add("Center", lcentre);
	this.getContentPane().add("South", l3);
    }


    
    void refresh() {
	String s = "Can't find any Nova System on this chain.";
	ids.setVisible(false);
	ids.removeAllItems();
	parent.setStandalone(true);

	if (novas.size() > 0) {
	    for (int e = 0; e < novas.size(); e++) {
		ids.addItem( (Integer)novas.elementAt(e) );
	    }
	    resultLabel.setText("Select the SysEx ID of your Nova System : ");
	    ids.setVisible(true);
	    parent.setStandalone(false);
	}
    }

    void lookUp() {
	novas.removeAllElements();
	resultLabel.setText("Can't find any Nova System on this chain.");
	ids.setVisible(false);
	ids.removeAllItems();
	send( inqMessage );

	/* debug
	novas.add(new Integer(1));
	refresh();
	*/
    }

    /* Called by
       -- MMJ MidiSystemListener method
       -- MMJ CheckBox is selected
     */
    public void systemChanged() {
	rebuildCombos();// abstract
	// Useless :
	// This is done via Lookup / rebuildCombos / addItem autoselect
	//parent.setNovas( novas );
	pack();
    }// End of systemChanged

	
    
    /* ItemChange Interface */
    public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange()==ItemEvent.SELECTED)
	    useMMJ=true;
	else useMMJ=false;

	systemChanged();
    }

    /* push input data into the queue
     */
    void processInputData(byte[] data) {
	try {
	    queue.offer( new MMidiMessage(data) );
	} catch (Exception e) {
	    System.out.println("Interruption while putting to the queue");
	}
    }

    /* process one message at once (call by Run)*/
    void processMidiMessage(MMidiMessage m) {
	parent.logSilent(m.toString());
	byte[]data = m.getMessage();
	int len = data.length;

	switch (len) {
	case 2:// Program Change
	    if (isPC(data)) parent.logSilent("= Program Change");
	    else parent.log("2 unknown bytes");
	    break;
	case 3:// Control Change
	    if (isCC(data)) parent.logSilent("= Control Change");
	    else parent.log("3 unknown bytes");
	    break;
	case 6:// Out of bounds preset
	    if (isPresetOutOfBounds(data)) parent.logSilent("= Preset empty/out of bounds");
	    break;
	case 17://meMova
	    if (isNova( data )) {
		parent.logSilent("= Nova identification message");
		Integer id = new Integer( data[2] );
		if (!novas.contains(id)) { 
		    novas.add( id );
		    //parent.setNovas( novas );
		    refresh();
		}
	    } else parent.logSilent("Not a Nova identification message");
	    break;

	case 520://Preset or Variation Dump
	    if (Patch.check(data)) {
		// Get n (Patch number) to discriminate Presets (< 91) and Variations (> 90)
		int n = data[8];
		//System.out.println("Preset number is "+n);

		if ((n>=0) && (n<=90)) {
		    parent.logSilent("= Preset");
		    Patch p = new Patch( parent, data );
		    parent.add(p);
		} else 
		    if ((n>90) && (n<=117)) {
			parent.logSilent("= Variation");
			Variation v = new Variation( parent, data );
			parent.add(v);
		} else 
			parent.log("Wrong Preset Number: "+n);
	    } else parent.log("Not a Preset");
	    break;
	case 526://System Dump
	    if (SystemDump.check(data)) {
		parent.logSilent("= System Dump");
		parent.add(new SystemDump( parent, data));
	    } else parent.log("Not a SystemDump");
	    break;
	default:
	    parent.log("Unknown message was received");
	}
	 parent.logSilent();
    }


    void send(int deviceID, byte[] data) {
	send( (byte) deviceID, data );
    }

    void send(byte deviceID, byte[] data) {
	int l = data.length;

	if (l==6) // Universal Device Inquiry REQUEST
	    data[2] = deviceID;
	else if (l==11) // Patch or System Request
	    data[4] = deviceID;
	
	send( data );// abstract
    }

    void sendPatchRequest(byte deviceID, byte num) {
	byte[] data = new byte[ reqPreset.length ];
	for (int i=0; i<data.length; i++) data[i]=reqPreset[i];
	data[4] = deviceID;
	data[8] = num;
	send( data );// abstract
    }

    void sendSystemDumpRequest(byte deviceID) {
	byte[] data = new byte[ reqSystem.length ];
	for (int i=0; i<data.length; i++) data[i]=reqSystem[i];
	data[4] = deviceID;
	send( data );// abstract
    }

    /* check whether a MIDI device is a Nova 
       look into its reply to a Universal Device Inquiry
     */
    static boolean isNova( byte[] data ) {
	boolean rep = true ;
	if (data.length != meNova.length) return false;

	for (int i=0; i < meNova.length ; i++) {
	    if ( i == 2 ) continue; //DeviceID may be different
	    //if ( i == 7 ) continue; //3rd byte of Manufacturer may differ
	    if ((i>8) && (i<16)) continue; // Software version does not matter
	    
	    if ( data[i] != meNova[i] ) {
		rep = false;
		break;
	    }
	}
	return rep;
    }

    Integer getCurrentSysExID() {
	if (ids.getItemCount() == 0) return null;
	else return (Integer) ids.getSelectedItem();
    }

    static boolean isPC( byte[] data) {
	if (data.length != 2) return false;
	MMidiMessage m = new MMidiMessage( data);
	if (m.getStatus() == 0xC) return true;
	else return false;
    }

    static boolean isCC( byte[] data) {
	if (data.length != 3) return false;
	MMidiMessage m = new MMidiMessage( data );
	if (m.getStatus() == 0xB) return true;
	else return false;
    }

    static boolean isPresetOutOfBounds( byte[] data ) {
	boolean rep = true ;
	if (data.length != oob.length) return false;

	for (int i=0; i < oob.length ; i++) {
	    if ( i == 2 ) continue; //DeviceID may be different
	    
	    if ( data[i] != oob[i] ) {
		rep = false;
		break;
	    }
	}
	return rep;
    }

    /* MidiInterface consumes the queue */
    public void run() {
	while (true) {
	    final MMidiMessage msg = queue.poll();
	    if (msg != null) processMidiMessage( msg );
	}
    }
}
			

